/* GUF Build - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <http://www.gauge3d.org/guf>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Binary.h"
#include "Compiler.h"
#include "Filesystem.h"
#include <iostream>
#include <fstream>

namespace gufbuild {
using namespace std;

// class member functions ============================================
Binary::Binary(string name, bool isLibrary, bool notGuf, string sourcefile, int linenum)
	: mName(name),
	  mIsLibrary(isLibrary),
	  mNotGuf(notGuf),
	  mMainObject(NULL),
	  mSourceFile(sourcefile),
	  mLineNum(linenum),
	  mBuilt(false),
	  mCleaned(false),
	  mChecked(false),
	  mDocumented(false)
{
}

Binary::~Binary()
{
}

Buildable::BuildResult Binary::Build()
{
	if(mBuilt)
		return mBuildResult;
	mBuilt = true;

	//TODO: build required libs

	//setup cflags and ldflags
	Compiler::ClearLibraries();
	Compiler::ClearParameters();
	Compiler::ClearLinkerParameters();

	Compiler::AddParameters(mCFlags);
	Compiler::AddLinkerParameters(mLdFlags);

	mBuildResult = SKIPPED; //assume skipped until we decide otherwise

	//build all parts.
	BuildResult childResult;
	for(unsigned int i = 0; i < mObjects.size(); i++)
	{
		childResult = mObjects[i]->Build();

		switch(childResult)
		{
		case FAILED:
			return mBuildResult = FAILED;
		case UPDATED:
			mBuildResult = UPDATED;
			break;
		case SKIPPED:
			break;
		}
	}

	string outputFile;
	vector<string > objects;
	GetObjects(objects);

	if(mIsLibrary)
		outputFile = Compiler::GetLibraryFileForName(mName);
	else
		outputFile = Compiler::GetExecutableFileForName(mName);

	if(mBuildResult == SKIPPED && Filesystem::Exists(outputFile))
	{
		//double-check whether we need to re-link or not
		bool needUpdate = false;
		unsigned int timestamp = Filesystem::GetTimestamp(outputFile);

		//check to see if any of the objects are newer than the output file.
		for(unsigned int i = 0; i < objects.size(); i++)
		{
			if(Filesystem::GetTimestamp(objects[i]) > timestamp)
			{
				//object file newer than linked output
				needUpdate = true;
				break;
			}
		}

		if(!needUpdate)
			return mBuildResult = SKIPPED;
	}

	string type;
	if(mIsLibrary)
		type = "library ";
	else
		type = "executable ";

	cout << "Linking " << type << mName << "...";
	if(Filesystem::IsStderrRedirected())
		cerr << "Linking " << type << mName << "..." << endl;
	else
		cout << endl;

	bool success;

	if(mIsLibrary)
		success = Compiler::LinkLibrary(objects, mName);
	else
	{
		string objectNameWithUnderscores = mMainObject->GetNameWithUnderscores();

		if(!mNotGuf)
		{
			//have to output *more* autogen source for the main function...
			string autogenFile = "build/" + mName + "__main.cpp";
			string autogenObject = "build/" + mName + "__main.o";
			fstream autogen(autogenFile.c_str(), ios::out);
			autogen << "extern \"C\" {\n";
			autogen << "   int " << objectNameWithUnderscores << "__Main(int argc, char* argv[]);\n";
			autogen << "}\n";
			autogen << "int main(int argc, char* argv[])\n";
			autogen << "   { return " << objectNameWithUnderscores << "__Main(argc, argv); }\n";
			autogen.close();

			if(!Compiler::Compile(autogenFile, autogenObject, false))
			{
				if(Filesystem::IsStderrRedirected())
				{
					cout << "(errors logged)" << endl;
					cout << endl;
					cout << "Errors occurred when compiling code generated by gufbuild which" << endl;
					cout << "contained the main() function.  This should never happen.  This" << endl;
					cout << "is probably a bug in gufbuild.  Please e-mail <temporal@gauge3d.org>." << endl;
				}

				cerr << endl;
				cerr << "Errors occurred when compiling code generated by gufbuild which" << endl;
				cerr << "contained the main() function.  This should never happen.  This" << endl;
				cerr << "is probably a bug in gufbuild.  Please e-mail <temporal@gauge3d.org>." << endl;
				return mBuildResult = FAILED;
			}
		}

//		objects.push_back(autogenObject);  //already added by GetObjects()
		success = Compiler::LinkExecutable(objects, mName);
	}

	if(Filesystem::IsStderrRedirected())
	{
		if(success)
			cout << endl;
		else
			cout << " (errors logged)" << endl;
		cerr << endl;
	}

	if(success)
		return mBuildResult = UPDATED;
//	else
		return mBuildResult = FAILED;
}

bool Binary::Check()
{
	if(mChecked)
		return mCheckResult;
	mChecked = true;

	mCheckResult = true;

	if(mMainObject == NULL)
	{
		cout << mSourceFile << ":" << mLineNum << ": ";
		if(mIsLibrary)
			cout << "Library ";
		else
			cout << "Executable ";
		cout << mName << " has no main package or module." << endl;
		mCheckResult = false;
	}

	for(unsigned int i = 0; i < mObjects.size(); i++)
	{
		if(!mObjects[i]->Check())
			mCheckResult = false;
	}

	//TODO: check headers and libs

	return mCheckResult;
}

bool Binary::Clean()
{
	if(mCleaned)
		return mCleanResult;
	mCleaned = true;

	cout << "Cleaning ";
	if(mIsLibrary)
		cout << "library ";
	else
		cout << "executable ";
	cout << mName << "..." << endl;

	string outputFile;

	if(mIsLibrary)
	{
		outputFile = Compiler::GetLibraryFileForName(mName);
		string dllFile = Compiler::GetDllFileForName(mName);
		if(dllFile != outputFile)
			Filesystem::Delete(dllFile);
	}
	else
		outputFile = Compiler::GetExecutableFileForName(mName);

	Filesystem::Delete(outputFile);

	if(!mIsLibrary)
	{
		Filesystem::Delete("build/" + mName + "__main.cpp");
		Filesystem::Delete("build/" + mName + "__main.o");
	}

	for(unsigned int i = 0; i < mObjects.size(); i++)
	{
		if(!mObjects[i]->Clean())
			return mCleanResult = false;
	}

	return mCleanResult = true;
}

bool Binary::Document()
{
	if(mDocumented)
		return mDocumentResult;
	mDocumented = true;

	//TODO

	return mDocumentResult = false;
}


void Binary::GetRequiredHeaders(vector<string >& list)
{
	for(unsigned int i = 0; i < mObjects.size(); i++)
		mObjects[i]->GetRequiredHeaders(list);
}

void Binary::GetRequiredLibraries(vector<string >& list)
{
	for(unsigned int i = 0; i < mObjects.size(); i++)
		mObjects[i]->GetRequiredLibraries(list);
}

void Binary::GetObjects(vector<string >& objects)
{
	for(unsigned int i = 0; i < mObjects.size(); i++)
		mObjects[i]->GetObjects(objects);
	if(!mIsLibrary && !mNotGuf)
		objects.push_back("build/" + mName + "__main.o");
}

string Binary::GetName() {return mName;}
string Binary::GetNameWithUnderscores() {return mName;}

void Binary::AddObject(Buildable* object, bool main)
{
	mObjects.push_back(object);
	if(main)
		mMainObject = object;
}

void Binary::AddCFlag(string cflag)
{
	mCFlags.push_back(cflag);
}

void Binary::AddLdFlag(string ldflag)
{
	mLdFlags.push_back(ldflag);
}

} //namespace gufbuild
