/* GUF Build - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <http://www.gauge3d.org/guf>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Package.h"
#include <iostream>
#include "Compiler.h"
#include "Filesystem.h"
#include "Module.h"

namespace gufbuild {
using namespace std;

// class member functions ============================================
Package::Package(string name, bool isPublic, string sourcefile, int linenum)
	: mIsPublic(isPublic),
	  mExecutable(false),
	  mTestable(false),
	  mMainModule(NULL),
	  mSourceFile(sourcefile),
	  mLineNum(linenum),
	  mBuilt(false),
	  mCleaned(false),
	  mChecked(false),
	  mDocumented(false)
{
	mName = mFilename = mNameWithUnderscores = name;

	unsigned int pos;
	while((pos = mFilename.find_first_of(':')) != string::npos)
	{
		mFilename.replace(pos, 2, "/");
		mNameWithUnderscores.replace(pos, 2, "_");
	}
}

Package::~Package()
{
	for(unsigned int i = 0; i < mBuildableList.size(); i++)
		delete mBuildableList[i];
}

Buildable::BuildResult Package::Build()
{
	if(mBuilt)
		return mBuildResult;  //already built!

	mBuilt = true;
	//assume skipped untill we actually build something...
	mBuildResult = SKIPPED;

	//whether or not we printed the "Compiling [whatever]..." message.
	bool alreadyPrintedMessage = false;

	//build all sub-packages and modules.
	BuildResult childResult;
	for(unsigned int i = 0; i < mBuildableList.size(); i++)
	{
		childResult = mBuildableList[i]->Build();

		switch(childResult)
		{
		case FAILED:
			mBuildResult = FAILED;
			break;
		case UPDATED:
			mBuildResult = UPDATED;
			break;
		case SKIPPED:
			break;
		}

		if(mBuildResult == FAILED)
			break;
	}

	if((mBuildResult != FAILED) && (mExecutable || mTestable))
	{
		//need to output some extra stuff to support the executable and testable flags.
		string autogenSource = "build/" + mNameWithUnderscores + "__autogen.cpp";
		string autogenObject = "build/" + mNameWithUnderscores + "__autogen.o";

		//make sure we really need to compile it...
		if(mBuildResult == UPDATED || !Filesystem::Exists(autogenObject))
		{
			//yep...
			cout << "Compiling " << mName << "...";
			if(Filesystem::IsStderrRedirected())
				cerr << "Compiling " << mName << "..." << endl;
			else
				cout << endl;

			alreadyPrintedMessage = true;

			if(mMainModule->CompileAutogenStuff(mNameWithUnderscores))
				mBuildResult = UPDATED;
			else
				mBuildResult = FAILED;
		}
	}

	if(Filesystem::IsStderrRedirected() && alreadyPrintedMessage)
	{
		cout << endl;
		cerr << endl;
	}

	return mBuildResult;
}

bool Package::Check()
{
	if(mChecked)
		return mCheckResult;
	mChecked = true;

	mCheckResult = true;

	for(unsigned int i = 0; i < mBuildableList.size(); i++)
	{
		if(!mBuildableList[i]->Check())
			mCheckResult = false;
	}

	if(mExecutable || mTestable)
	{
		if(mMainModule == NULL)
		{
			cerr << mSourceFile << ":" << mLineNum << ": Package " << mName
			     << " set executable or testable but no main module was given." << endl;
			mCheckResult = false;
		}
	}

	return mCheckResult;
}

bool Package::Clean()
{
	if(mCleaned)
		return mCleanResult;
	mCleaned = true;

	cout << "Cleaning " << mName << "..." << endl;
	if(mExecutable || mTestable)
	{
		Filesystem::Delete("build/" + mNameWithUnderscores + "__autogen.cpp");
		Filesystem::Delete("build/" + mNameWithUnderscores + "__autogen.o");
	}

	for(unsigned int i = 0; i < mBuildableList.size(); i++)
	{
		if(!mBuildableList[i]->Clean())
			return mCleanResult = false;
	}

	return mCleanResult = true;
}

bool Package::Document()
{
	if(mDocumented)
		return mDocumentResult;
	mDocumented = true;

	//TODO

	return mDocumentResult = false;
}

void Package::AddRequiredHeader(string header)
	{ mRequiredHeaders.push_back(header); }

void Package::AddRequiredLibrary(string library)
	{ mRequiredLibraries.push_back(library); }

void Package::GetRequiredHeaders(vector<string >& list)
{
	list.insert(list.end(), mRequiredHeaders.begin(), mRequiredHeaders.end());
	for(unsigned int i = 0; i < mBuildableList.size(); i++)
		mBuildableList[i]->GetRequiredHeaders(list);
}

void Package::GetRequiredLibraries(vector<string >& list)
{
	list.insert(list.end(), mRequiredLibraries.begin(), mRequiredLibraries.end());
	for(unsigned int i = 0; i < mBuildableList.size(); i++)
		mBuildableList[i]->GetRequiredLibraries(list);
}

void Package::GetObjects(vector<string >& objects)
{
	for(unsigned int i = 0; i < mBuildableList.size(); i++)
		mBuildableList[i]->GetObjects(objects);
	if(mExecutable || mTestable)
		objects.push_back("build/" + mNameWithUnderscores + "__autogen.o");
}

string Package::GetName() {return mName;}
string Package::GetNameWithUnderscores() {return mNameWithUnderscores;}

Package* Package::AddSubPackage(string name, bool isPublic, string sourceFile, int lineNum)
{
	unsigned int pos = name.find_first_of(':');
	if(!mIsPublic)
		isPublic = false;
	if(pos == string::npos)
	{
		map<string, Buildable* >::iterator iter = mBuildableMap.find(name);
		if(iter == mBuildableMap.end())
		{
			Package* package;
			if(mName.length() > 0)
				package = new Package(mName + "::" + name, isPublic, sourceFile, lineNum);
			else
				package = new Package(name, isPublic, sourceFile, lineNum);

			mBuildableMap[name] = package;
			mBuildableList.push_back(package);
			return package;
		}
		else
			return dynamic_cast<Package* >(iter->second);
	}
	else
	{
		string nextPackage = name.substr(0, pos);
		name = name.substr(pos + 2);
		map<string, Buildable* >::iterator iter = mBuildableMap.find(nextPackage);
		if(iter == mBuildableMap.end())
			return (AddSubPackage(nextPackage, true, sourceFile, lineNum)
			      ->AddSubPackage(name, isPublic, sourceFile, lineNum));
		else if(Package* package = dynamic_cast<Package* >(iter->second))
			return package->AddSubPackage(name, isPublic, sourceFile, lineNum);
		else
			return NULL;
	}
}

Module* Package::AddModule(string name, bool isPublic, bool mainModule, string sourceFile, int lineNum)
{
	unsigned int pos = name.find_last_of(':');
	if(!mIsPublic)
		isPublic = false;
	if(pos == string::npos)
	{
		map<string, Buildable* >::iterator iter = mBuildableMap.find(name);
		if(iter == mBuildableMap.end())
		{
			Module* module;
			if(mName.length() > 0)
				module = new Module(mName + "::" + name, isPublic, sourceFile, lineNum);
			else
				module = new Module(name, isPublic, sourceFile, lineNum);

			if(mainModule)
				mMainModule = module;
			mBuildableMap[name] = module;
			mBuildableList.push_back(module);
			return module;
		}
		else
		{
			Module* module = dynamic_cast<Module* >(iter->second);
			if(mainModule && module != NULL)
				mMainModule = module;
			return module;
		}
	}
	else
	{
		string packageName = name.substr(0, pos);
		name = name.substr(pos + 2);
		Package* package = AddSubPackage(packageName, true, sourceFile, lineNum);
		if(package != NULL)
			return package->AddModule(name, isPublic, mainModule, sourceFile, lineNum);
		else
			return NULL;
	}
}

Buildable* Package::GetBuildable(string name)
{
	int pos = name.find_first_of(':');
	if(pos == -1)
	{
		map<string, Buildable* >::iterator iter = mBuildableMap.find(name);
		if(iter == mBuildableMap.end())
			return NULL;
		else
			return iter->second;
	}
	else
	{
		string nextPackage = name.substr(0, pos);
		name = name.substr(pos + 2);
		map<string, Buildable* >::iterator iter = mBuildableMap.find(nextPackage);
		if(iter == mBuildableMap.end())
			return NULL;
		else if(Package* package = dynamic_cast<Package* >(iter->second))
			return package->GetBuildable(name);
		else
			return NULL;
	}
}

} //namespace gufbuild
